from bs4 import BeautifulSoup


class Acidente:
    """ Classe responsável por manter atributos referentes a acidentes.

    Atributos
    ---------
    url: str
        url acessada para obter informações sobre acidentes
    links: list
        lista de códigos que formam o link para download.
    """

    def __init__(self):
        self.url = 'https://portal.prf.gov.br/dados-abertos-acidentes'
        self._tipo = ''
        self.links = {}

    def carregar_links(self, response, tipo):
        """ Carrega links disponíveis na página de infrações. """
        # Recupera o HTML
        html_data = response.text
        # Cria um objeto BeautifulSoup a partir do HTML
        soup = BeautifulSoup(html_data, features="html.parser")

        # Recupera título
        titulos = soup.findAll("h2")

        links = []

        # Busca links
        for i, tag in enumerate(titulos):
            titulo = tag.text
            # Checa título de H2, procura lista de links próximo
            if titulo == tipo:
                links = tag.find_next_sibling('ul').findAll('a')
                break

        # Filtra links referentes aos anos
        self.links[tipo] = {int(link.text.split('-')[0]): link['href'].split('/')[-1]
                            for link in links if link.text.startswith('20')}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `prf_api` package."""

import pytest

from click.testing import CliRunner

from prf_api import prf_api
from prf_api import cli
import os


@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_init():
    """Sample pytest test function with the pytest fixture as an argument."""
    prf_data = prf_api.PRFApi()
    prf_data.baixar('infracoes', anos=[2018, 2019])
    caminho = os.getcwd() + "/infracoes"
    assert os.path.exists(caminho)
    assert os.path.exists(caminho + "/2018")
    for mes in ['abr', 'ago', 'dez', 'fev', 'jan', 'jul',
                'jun', 'mai', 'mar', 'nov', 'out', 'set']:
        assert os.path.exists(caminho + "/2018/" + mes + ".csv")

    prf_data.baixar('acidentes_ocorrencia', anos=[2016])
    caminho = os.getcwd() + "/acidentes_ocorrencia"
    assert os.path.exists(caminho)
    assert os.path.exists(caminho + "/2016")


def test_dataframe():
    """ Testa criação de dataframe"""
    prf_data = prf_api.PRFApi()
    df = prf_data.dataframe('infracoes', anos=list(range(2017, 2019)),
                            estado='RN')
    assert df.uf_infracao.unique() == ['RN']

    anos = df.dat_infracao.str.split('-', n=1, expand=True)[0].unique()
    assert sorted(anos) == ['2017', '2018']
    assert len(df.columns) == 22

    df_acidente = prf_data.dataframe('acidentes_ocorrencia', anos=[2016],
                                     estado='RN')
    assert df_acidente.uf.unique() == ['RN']


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.main)
    assert result.exit_code == 0
    assert 'prf_api.cli.main' in result.output
    help_result = runner.invoke(cli.main, ['--help'])
    assert help_result.exit_code == 0
    assert '--help  Show this message and exit.' in help_result.output



# -*- coding: utf-8 -*-

import requests
import os
from .infracoes import Infracoes
from .acidentes import Acidente
from .arquivo import extrair_arquivos
import pandas as pd


class PRFApi:
    """ Classe principal usada para manipular dados abertos da Polícia
    Rodoviária Federal.
    """

    TIPOS = {'infracoes': 'infracoes',
             'acidentes_pessoa': 'Agrupados por pessoa',
             'acidentes_ocorrencia': 'Agrupados por ocorrência',
             'acidentes_agrupados': 'Agrupados por pessoa - \
Todas as causas e tipos de acidentes (a partir de 2017)'}

    # Mapeamento de coluna que deve ser
    # filtrada para cada localidade e cada
    # tipo de conjunto
    COLUNAS = {
        'infracoes': {
            'estado': 'uf_infracao'
        },
        'acidentes_pessoa': {'estado': 'uf'},
        'acidentes_ocorrencia': {'estado': 'uf'},
        'acidentes_agrupados': {'estado': 'uf'},
    }

    REGIOES = {
        'CO': ['DF', 'GO', 'MS', 'MT'],
        'N': ['AC', 'AP', 'AM', 'PA', 'RO', 'RR', 'TO'],
        'NE': ['AL', 'BA', 'CE', 'MA', 'PB', 'PE', 'PI', 'RN', 'SE'],
        'S': ['PR', 'RS', 'SC'],
        'SE': ['ES', 'MG', 'RJ', 'SP']
    }

    def __init__(self):
        self.url = 'https://www.prf.gov.br/portal/dados-abertos-infracoes'
        self.download_url = 'http://arquivos.prf.gov.br/arquivos/index.php/s/{}/download'

        self._carregar_links()

    def _exibir_erro(self, msg: str, ex: Exception = None):
        """Exibir mensagem personalizada para erro

        Parâmetros
        ----------
        ex: Exception
            exceção que foi detectada
        msg: str
            mensagem personalizada para erro
        """
        print('\033[91m{}\033[0m'.format(msg))
        if ex is not None:
            print(ex)
        print()

    def _carregar_links(self):
        """Função que busca o link para cada infração"""
        self.infracoes = Infracoes()
        self.acidentes = Acidente()
        try:
            response_infracoes = requests.get(self.infracoes.url)
            response_acidentes = requests.get(self.acidentes.url)
        except requests.exceptions.ConnectionError as ex:
            self._exibir_erro("Falha de conexão. "
                              "Verifique sua conexão e tente novamente.", ex)
            return
        except Exception as ex:
            self._exibir_erro("Ocorreu um erro inesperado. "
                              "Verifique sua conexão e tente novamente.", ex)
            return

        self.infracoes.carregar_links(response_infracoes)
        for tipo in self.TIPOS.values():
            if tipo is not "infracoes":
                self.acidentes.carregar_links(response_acidentes, tipo)

    def _criar_diretorio(self, caminho: str) -> str:
        """Cria o diretório, caso ele não exista.
        Parâmetros
        ----------
        path: str
            o caminho da pasta onde serão adicionados os arquivos."""
        if not os.path.exists(caminho):
            os.makedirs(caminho)

        return caminho

    def _filtrar_dataframe(self, df: pd.DataFrame, coluna: str,
                           area: {str, list}) -> pd.DataFrame:
        """ Filtra dataframe de acordo com estado ou área.

        Parâmetros
        ----------
        df: pd.DataFrame
            dataframe sobre o qual se aplica a filtragem
        coluna: str
            coluna sobre a qual se aplica o filtro
        area: str  ou list
            estado (sigla) ou regiao (lista de estados)
        """
        if type(area) == str:
            query = " {} == '{}'"
        elif type(area) == list:
            query = "{} in {}"
        else:
            self._exibir_erro("Área {} não disponível".format(area))

        return df.query(query.format(coluna, area))

    def baixar(self, tipo: {'infracoes', 'acidentes_pessoa',
               'acidentes_ocorrencia', 'acidentes_agrupados'},
               anos: list = None, caminho: str = os.getcwd()):
        """Realiza o download dos conjuntos de dados de acordo com o
        tipo desejado

        Parâmetros
        ----------
        tipo : {'infracoes', 'acidentes_pessoa', 'acidentes_ocorrencia',
                'acidentes_agrupados'}
            tipo de dados que se deseja realizar o download (por padrão,
            'infracoes')
        caminho: str
            o caminho da pasta onde serão adicionados os arquivos
            (por padrão, a pasta atual).
        anos: list
            lista de anos dos dados
        """
        if tipo not in self.TIPOS.keys():
            self._exibir_erro("Tipo '{}' é inválido. ".format(tipo))
            return

        if tipo == "infracoes":
            links = self.infracoes.links
        else:
            links = self.acidentes.links[self.TIPOS[tipo]]

        # Se não for passada uma lista, baixar todo os anos
        if anos is None:
            anos = links.keys()

        # Checa se os anos são válidos
        for ano in anos:
            if not (ano in links.keys()):
                self._exibir_erro("Ano {} não disponível".format(ano))

        # Realiza o download
        caminho = self._criar_diretorio('{}/{}'.format(caminho, tipo))
        for ano in anos:
            try:
                link = self.download_url.format(links[ano])
                print("Buscando datasets de {} para o ano {}...".format(tipo,
                                                                        ano))
                dataset = requests.get(link)
                diretorio = self._criar_diretorio('{}/{}'.format(caminho, ano))

                # Carrega e descompacta arquivo comprimido
                formato_arquivo = dataset.headers['Content-Type'].split('/')[-1]
                extrair_arquivos(formato_arquivo, diretorio, dataset.content)
            except requests.exceptions.ConnectionError as ex:
                self._exibir_erro("Falha de conexão. "
                                  "Verifique sua conexão e tente novamente.", ex)
            except Exception as ex:
                self._exibir_erro("Ocorreu um erro inesperado. "
                                  "Verifique sua conexão e tente novamente.", ex)            

    def dataframe(self, tipo: {'infracoes', 'acidentes_pessoa',
                  'acidentes_ocorrencia', 'acidentes_agrupados'},
                  anos: list = None, caminho: str = os.getcwd(),
                  estado: str = None,
                  regiao: {'CO', 'N', 'NE', 'S', 'SE'}=None) -> pd.DataFrame:
        """Trasforma os csvs em dataframes.

        Parâmetros
        ----------
        tipo : {'infracoes', 'acidentes_pessoa', 'acidentes_ocorrencia',
                'acidentes_agrupados'}
            tipo de dados que se deseja realizar o download (por padrão,
            'infracoes')
        caminho: str
            o caminho da pasta onde serão adicionados os arquivos
            (por padrão, a pasta atual).
        anos: list
            lista de anos dos dados.
        estado: str
            sigla do estado que servirá para a filtragem.
        regiao: {'CO','N', 'NE', 'S', 'SE'}
            sigla da região cujos estados serão filtrados.

        Retorno
        -------
        dataframe com todos os dados.
        """
        # Verifica se tipo é válido
        if tipo not in self.TIPOS.keys():
            self._exibir_erro("Tipo '{}' é inválido. ".format(tipo))
            return

        if tipo == "infracoes":
            links = self.infracoes.links
        else:
            links = self.acidentes.links[self.TIPOS[tipo]]

        # Se não for passada uma lista, baixar todo os anos
        if anos is None:
            anos = links.keys()

        # Verifica se estado é válido
        estados = [x for regiao in list(self.REGIOES.values()) for x in regiao]
        if (estado is not None) and (estado not in estados):
            self._exibir_erro("Estado '{}' é inválido. ".format(estado))
            return

        # Verifica se região é válida
        if (regiao is not None) and (regiao not in self.REGIOES.keys()):
            self._exibir_erro("Região '{}' é inválida. ".format(regiao))
            return

        # Verifica se todos os anos foram baixados
        for ano in anos:
            pasta = '{}/{}/{}'.format(caminho, tipo, ano)
            if not os.path.exists(pasta):
                # Se não estiver baixado, realizar o download
                self.baixar(tipo, [ano], caminho)

        # Recupera csvs e cria dataframe
        dataframe = pd.DataFrame()  # Cria df vazio
        for ano in anos:
            print(">>> Criando dataframe com ano {}...".format(ano))
            pasta = '{}/{}/{}'.format(caminho, tipo, ano)
            for arquivo in os.listdir(pasta):
                caminho_arquivo = '{}/{}'.format(pasta, arquivo)
                try:
                    df = pd.read_csv(caminho_arquivo, encoding='latin1')
                except pd.errors.ParserError:
                    # Checa se o csv foi separado corretamente
                    df = pd.read_csv(caminho_arquivo, encoding='latin1',
                                     sep=';')

                # Checa se o csv foi separado corretamente
                if len(df.columns) == 1:
                    df = pd.read_csv(caminho_arquivo, encoding='latin1',
                                     sep=';')

                coluna = self.COLUNAS[tipo]['estado']
                # Realiza a filtragem, se algum parâmetro de busca for usado
                if estado is not None:
                    df = self._filtrar_dataframe(df, coluna, estado)
                elif regiao is not None:
                    # Filtra todos os estados da região
                    df = self._filtrar_dataframe(df, coluna,
                                                 self.REGIOES[regiao])
                dataframe = dataframe.append(df, ignore_index=True)

                del df  # Libera memória

        return dataframe



from io import BytesIO
from zipfile import ZipFile
from rarfile import RarFile
import os


def extrair_rar(rf: RarFile, caminho: str):
    """ Extrai csvs de um arquivo . rar.

    Parâmetros
    ----------
    rf: Rarfile
        conteúdo do arquivo compactado.
    caminho: str
        caminho para pasta onde os arquivos devem ser salvos.
    """
    ano = caminho.split('/')[-1]
    n_arquivos = len(rf.infolist())
    for f in rf.infolist():
        # Filtra arquivos csvs comprimidos
        if f.filename.endswith('csv'):
            filename = f.filename.split('/')[-1]
            print("\033[94m>> Baixando {}/{}\033[0m".format(ano, filename))
            with open(caminho + '/' + filename, "wb") as of:
                of.write(rf.read(f.filename))


def extrair_zip(zp: ZipFile, caminho: str):
    """ Extrai csvs de um arquivo .zip.

    Parâmetros
    ----------
    zp: ZipFile
        conteúdo do arquivo compactado.
    caminho: str
        caminho para pasta onde os arquivos devem ser salvos.
    """
    ano = caminho.split('/')[-1]
    for f in zp.namelist():
        if f.endswith('csv'):
            filename = f.split('/')[-1]
            print("\033[94m>> Baixando {}/{}\033[0m".format(ano, filename))
            with open(caminho + '/' + filename, "wb") as of:
                of.write(zp.read(f))


def extrair_arquivos(tipo: str, caminho: str, conteudo):
    """ Extrai csvs de um arquivo compactado.

    Parâmetros
    ----------
    tipo: str
        tipo do arquivo: rar ou zip.
    caminho: str
        caminho para pasta onde os arquivos devem ser salvos.
    conteudo
        bytes do arquivo compactado.
    """
    if tipo == "x-rar-compressed":
        # Salva temporariamente o .rar
        with open('file.rar', 'wb') as f:
            f.write(conteudo)
        # Extrai arquivos csv
        with RarFile("file.rar") as rf:
            extrair_rar(rf, caminho)

        # Apaga .rar
        os.remove('file.rar')
    else:
        zipfile = ZipFile(BytesIO(conteudo))
        with zipfile as zp:
            extrair_zip(zp, caminho)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['Click>=6.0', 'requests==2.22.0',
                'pandas==0.24.2', 'rarfile==3.0',
                'beautifulsoup4==4.7.1', 'unrar==0.3']

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    author="Mayra Azevedo",
    author_email='mayradazevedo@ufrn.edu.br',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: Portuguese',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="API de dados abertos da PRF",
    entry_points={
        'console_scripts': [
            'prf_api=prf_api.cli:main',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords=['prf', 'wrapper'],
    name='prf_api',
    packages=find_packages(include=['prf_api']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/nymarya/prf_api',
    version='0.2.2',
    zip_safe=False,
)



from bs4 import BeautifulSoup


class Infracoes:
    """ Classe responsável por manter atributos referentes a infrações.

    Atributos
    ---------
    url: str
        url acessada para obter informações sobre infrações
    links: list
        lista de códigos que formam o link para download.
    """

    def __init__(self):
        self.url = 'https://portal.prf.gov.br/dados-abertos-infracoes'

    def carregar_links(self, response):
        """ Carrega links disponíveis na página de infrações. """
        # Recupera o HTML
        html_data = response.text
        # Cria um objeto BeautifulSoup a partir do HTML
        soup = BeautifulSoup(html_data, features="html.parser")

        links = soup.findAll("a")
        # Filtra links referentes aos anos
        self.links = {int(link.text): link['href'].split('/')[-1]
                      for link in links if link.text.startswith('20')}
